import sys
import os
from typing import Literal
from pathlib import Path
import pytest
from peakrdl.main import main
from mnrs_exporter import ROOT_DIR, TEMPLATE_DIR

TEST_DIR = Path(os.path.dirname(__file__))
path_prefix = os.path.relpath(TEST_DIR, start=TEMPLATE_DIR)

gen_option_output_map = {"addrmap": "testing_addrmap.h",
                         "regfile": ["gen/array_test_regs.h",
                                     "gen/different_size_test_regs.h",
                                     "gen/read_write_test_regs.h",
                                     "gen/single_field_test_regs.h"],
                         "module_header": ["array_test.h",
                                           "different_size_test.h",
                                           "read_write_test.h",
                                           "single_field_test.h"],
                         "module_source": ["array_test.cpp",
                                           "different_size_test.cpp",
                                           "read_write_test.cpp",
                                           "single_field_test.cpp"]
                         }


def run_mnrs_exporter(output_dir: Path,
                      gen_option: Literal["addrmap", "regfile", "module_header", "module_source"],
                      template_name: Literal["TestAddrMap.jinja",
                                             "TestRegfile.jinja"],
                      overwrite: bool = False
                      ):
    arg_list = ["peakrdl", "mnrs", f"--{gen_option}", str(TEST_DIR/"test_addrmap.rdl"),
                "-o", str(output_dir), "--template", template_name]
    if overwrite:
        arg_list.append("--overwrite")
    sys.argv = arg_list
    main()


def check_output(output_dir: Path, gen_option: Literal["addrmap", "regfile", "module_header", "module_source"]):
    output = gen_option_output_map[gen_option]
    for file in output if isinstance(output, list) else [output]:
        with open(output_dir/file, encoding="utf8") as generated_file, \
                open(TEST_DIR/"references"/file, encoding="utf8") as reference_file:
            gen_content = generated_file.read()
            ref_content = reference_file.read()
            assert gen_content == ref_content


class TestGeneration():
    def test_addrmap_gen(self, tmp_path):
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{path_prefix}/TestAddrMap.jinja")
        check_output(tmp_path, "addrmap")

    def test_regfile_gen(self, tmp_path):
        run_mnrs_exporter(tmp_path, "regfile",
                          f"{path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "regfile")

    @pytest.mark.filterwarnings('ignore: Using "module_headers" and "template"')
    def test_module_header_gen(self, tmp_path):
        run_mnrs_exporter(tmp_path, "module_header",
                          f"{path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "module_header")

    @pytest.mark.filterwarnings('ignore: Using "module_sources" and "template"')
    def test_module_source_gen(self, tmp_path):
        run_mnrs_exporter(tmp_path, "module_source",
                          f"{path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "module_source")


class TestInput():
    def test_invalid_input(self):
        sys.argv = ["peakrdl", "mnrs", "--all",
                    str(TEST_DIR/"invalid_addrmap.rdl"), "-o", "remove_test_dir_should_not_exist"]
        with pytest.raises(ValueError):
            main()

    def test_realtive_path_template(self, tmp_path):
        cur_dir = os.getcwd()
        os.chdir(ROOT_DIR)
        run_mnrs_exporter(tmp_path, "addrmap", "test/TestAddrMap.jinja")
        check_output(tmp_path, "addrmap")
        os.chdir(cur_dir)

    def test_absolute_path_template(self, tmp_path):
        run_mnrs_exporter(tmp_path, "addrmap",
                          "/Users/alex/git/peakrdl_mnrs/test/TestAddrMap.jinja")
        check_output(tmp_path, "addrmap")


class TestOverwrite():
    def test_overwrite(self, tmp_path):
        testfile = f"{tmp_path}/testing_addrmap.h"
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{path_prefix}/TestAddrMap.jinja")

        mtime = os.path.getmtime(testfile)
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{path_prefix}/TestAddrMap.jinja")
        assert mtime == os.path.getmtime(testfile)

        run_mnrs_exporter(tmp_path, "addrmap", f"{path_prefix}/TestAddrMap.jinja",
                          overwrite=True)
        assert mtime != os.path.getmtime(testfile)

    @pytest.mark.filterwarnings('ignore: Using "module_headers" and "template"')
    def test_overwrite_protection(self, tmp_path):
        testfiles = [
            f"{tmp_path}/{file}" for file in gen_option_output_map["module_header"]]
        run_mnrs_exporter(tmp_path, "module_header",
                          f"{path_prefix}/TestRegfile.jinja")

        mtimes = [os.path.getmtime(testfile) for testfile in testfiles]
        run_mnrs_exporter(tmp_path, "module_header", f"{path_prefix}/TestRegfile.jinja",
                          overwrite=True)
        assert mtimes == [os.path.getmtime(testfile) for testfile in testfiles]
